from pathlib import Path
import parsy

from .parse import make_parser
from .util import flatten
from .check import check_files
from .issue import Issue, IssueLocation
from .file import File


def get_parsy_error_location(error, file_path):
    line, column = parsy.line_info_at(error.stream, error.index)
    return IssueLocation(
        line=line,
        column=column,
        file_path=file_path,
    )


def resolve_file_paths_(input_name, extensions):
    path = Path(input_name)
    if not path.exists():
        raise Exception('{} does not exist'.format(path))

    if path.is_dir():
        return flatten(
            resolve_file_paths_(child, extensions) for child in path.iterdir()
        )

    if not path.is_file():
        raise Exception('{} is not a regular file'.format(path))

    return [path] if path.suffix in extensions else []


def resolve_file_paths(input_names, extensions):
    path_lists = (resolve_file_paths_(i, extensions) for i in input_names)
    return flatten(path_lists)


def parse_file(path_and_config):
    """
    Returns a tuple ([Issue], File | None).
    """
    path, config = path_and_config

    with path.open('r') as f:
        source = f.read()

    parser = make_parser(config)

    try:
        file = File(
            path=path,
            source=source,
            lines=source.split('\n'),
            tree=parser['content'].parse(source),
        )
        return [], file
    except parsy.ParseError as error:
        location = get_parsy_error_location(error, path)
        issue = Issue(location, 'Parse error: ' + str(error))
        return [issue], None


def lint(paths, config):
    issues = []
    files = []

    from multiprocessing import Pool
    pool = Pool()

    parse_file_args = ((p, config) for p in paths)
    results = pool.map(parse_file, parse_file_args)
    for result in results:
        parse_issues, file = result
        issues += parse_issues
        if file is not None:
            files.append(file)

    if not config.get('parse_only', False):
        issues += check_files(files, config)

    return issues



import parsy as P

from .parse import (
    tag_name, tag_name_char, comment, jinja_comment,
    make_attribute_value_parser, make_attribute_parser,
    make_attributes_parser,
    make_closing_tag_parser, make_opening_tag_parser, make_parser,
)

from . import ast
from .ast import Location


parser = make_parser()
element = parser['element']
jinja = parser['jinja']
content = parser['content']
attribute_value = make_attribute_value_parser(jinja=jinja)
attribute = make_attribute_parser(jinja=jinja)
opening_tag = make_opening_tag_parser({}, jinja=jinja)


class DummyLocation():
    """
    Any instance of this class is equal to any Location.
    """
    def __eq__(self, other):
        if not isinstance(other, Location):
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return self.__class__.__name__ + '()'


def with_dummy_locations(node_class):
    def create_node(*args, **kwargs):
        kwargs = kwargs.copy()
        kwargs['begin'] = DummyLocation()
        kwargs['end'] = DummyLocation()

        try:
            return node_class(
                *args,
                **kwargs,
            )
        except Exception as error:
            print(node_class)
            raise error

    return create_node


Attribute = with_dummy_locations(ast.Attribute)
Element = with_dummy_locations(ast.Element)
ClosingTag = with_dummy_locations(ast.ClosingTag)
Comment = with_dummy_locations(ast.Comment)
JinjaComment = with_dummy_locations(ast.JinjaComment)
Integer = with_dummy_locations(ast.Integer)
Interp = with_dummy_locations(ast.Interpolated)
JinjaComment = with_dummy_locations(ast.JinjaComment)
JinjaElement = with_dummy_locations(ast.JinjaElement)
JinjaElementPart = with_dummy_locations(ast.JinjaElementPart)
JinjaTag = with_dummy_locations(ast.JinjaTag)
JinjaVariable = with_dummy_locations(ast.JinjaVariable)
OpeningTag = with_dummy_locations(ast.OpeningTag)
String = with_dummy_locations(ast.String)


def test_dummy_location():
    dummy = DummyLocation()
    real = Location(0, 0, 0)
    assert dummy == real
    assert real == dummy

    assert not dummy != real
    assert not real != dummy


def test_tag_name():
    assert tag_name_char.parse('a') == 'a'
    assert tag_name.parse('bcd-ef9') == 'bcd-ef9'


def test_attribute_value():
    assert attribute_value.parse('hello-world') == String(
        value=Interp('hello-world'),
        quote=None,
    )

    assert attribute_value.parse('hello{{a}}world') == String(
        value=Interp([
            'hello',
            JinjaVariable(content='a'),
            'world',
        ]),
        quote=None,
    )

    assert attribute_value.parse('123') == Integer(
        value=123,
        has_percent=False,
    )

    assert attribute_value.parse('"hello"') == String(
        value=Interp('hello'),
        quote='"',
    )

    assert attribute_value.parse("'hello'") == String(
        value=Interp('hello'),
        quote="'",
    )

    assert attribute_value.parse("''") == String(
        value=Interp([]),
        quote="'",
    )

    assert attribute_value.parse("'hello{{b}}world'") == String(
        value=Interp([
            'hello',
            JinjaVariable(content='b'),
            'world',
        ]),
        quote="'",
    )


def test_attribute():
    assert attribute.parse('hello=world') == Attribute(
        name=Interp('hello'),
        value=Interp(
            String(
                value=Interp('world'),
                quote=None,
            ),
        ),
    )

    assert attribute.parse('a= "b"') == Attribute(
        name=Interp('a'),
        value=Interp(
            String(
                value=Interp('b'),
                quote='"',
            ),
        ),
    )

    assert attribute.parse('a =b_c23') == Attribute(
        name=Interp('a'),
        value=Interp(
            String(
                value=Interp('b_c23'),
                quote=None,
            ),
        ),
    )

    assert attribute.parse('valueless-attribute') == Attribute(
        name=Interp('valueless-attribute'),
        value=None,
    )


def test_comment():
    assert comment.parse('<!--hello--world-->') == Comment(
        text='hello--world',
    )


def test_jinja_comment():
    assert jinja_comment.parse('{# hello world #}') == JinjaComment(
        text='hello world',
    )


def test_opening_tag():
    assert opening_tag.parse('<div>') == OpeningTag(
        name='div',
        attributes=Interp([]),
    )

    assert opening_tag.parse('<div\n >') == OpeningTag(
        name='div',
        attributes=Interp([]),
    )

    assert opening_tag.parse('<div class="red" style="" >') == OpeningTag(
        name='div',
        attributes=Interp([
            Attribute(
                name=Interp('class'),
                value=Interp(
                    String(
                        value=Interp('red'),
                        quote='"',
                    ),
                ),
            ),

            Attribute(
                name=Interp('style'),
                value=Interp(
                    String(
                        value=Interp([]),
                        quote='"',
                    ),
                ),
            ),
        ]),
    )


def test_closing_tag():
    closing_tag = make_closing_tag_parser(P.string('div'))
    assert closing_tag.parse('</div>') == ClosingTag(
        name='div',
    )


def test_raw_text_elements():
    assert element.parse('<style a=b> <wont-be-parsed> </style>') == Element(
        content=' <wont-be-parsed> ',

        opening_tag=OpeningTag(
            name='style',
            attributes=Interp([

                Attribute(
                    name=Interp('a'),
                    value=Interp(
                        String(
                            value=Interp('b'),
                            quote=None,
                        ),
                    ),
                ),

            ]),
        ),

        closing_tag=ClosingTag(
            name='style',
        ),
    )


def test_element():
    assert element.parse('<div> hey </div>') == Element(
        opening_tag=OpeningTag(
            name='div',
            attributes=Interp([]),
        ),
        content=Interp([' hey ']),
        closing_tag=ClosingTag(
            name='div',
        ),
    )

    attributes = [
        Attribute(
            name=Interp('onclick'),
            value=Interp(
                String(
                    value=Interp([]),
                    quote='"',
                ),
            ),
        ),

        JinjaVariable(content='var'),

        Attribute(
            name=Interp('class'),
            value=Interp(
                String(
                    value=Interp('red'),
                    quote='"',
                ),
            ),
        ),
    ]

    assert element.parse('<br onclick="" {{var}} class="red">') == Element(
        opening_tag=OpeningTag(
            name='br',
            attributes=Interp(attributes),
        ),
        closing_tag=None,
        content=None,
    )

    src = '<{% if a %}bcd{% endif %}></{% if a %}bcd{% endif %}>'
    assert src == str(element.parse(src))

    src = '<div{{ "ider" }}></div>'
    assert '<div {{ "ider" }}></div>' == str(element.parse(src))

    src = '<div{% if a %}foo="bar"{% endif %}></div>'
    assert '<div {% if a %}foo="bar"{% endif %}></div>' == \
        str(element.parse(src))

    src = '<div{% if a %} foo="bar"  a=2 {% endif %}></div>'
    assert '<div {% if a %}foo="bar"a=2{% endif %}></div>' == \
        str(element.parse(src))

    src = '<colgroup></colgroup>'
    assert src == str(element.parse(src))


def test_self_closing_elements():
    assert element.parse('<br>') == Element(
        opening_tag=OpeningTag(
            name='br',
            attributes=Interp([]),
        ),
        content=None,
        closing_tag=None,
    )

    src = '<br />'
    assert src == str(element.parse(src))


def test_jinja_blocks():
    assert jinja.parse('{% name something == 123 %}') == JinjaElement(
        parts=[
            JinjaElementPart(
                tag=JinjaTag(
                    name='name',
                    content='something == 123',
                ),
                content=None,
            ),
        ],
        closing_tag=None,
    )

    assert jinja.parse('{% if a %}b{% else %}c{% endif %}') == JinjaElement(
        parts=[
            JinjaElementPart(
                tag=JinjaTag(
                    name='if',
                    content='a',
                ),
                content=Interp(['b']),
            ),
            JinjaElementPart(
                tag=JinjaTag(
                    name='else',
                    content='',
                ),
                content=Interp(['c']),
            ),
        ],
        closing_tag=JinjaTag(
            name='endif',
            content='',
        ),
    )

    src = '{% if a %}b{% elif %}c{% elif %}d{% else %}e{% endif %}'
    assert src == str(jinja.parse(src))


def test_jinja_whitespace_controls():
    assert jinja.parse('{%- foo -%}') == JinjaElement(
        parts=[
            JinjaElementPart(
                tag=JinjaTag(
                    name='foo',
                    content='',
                    left_minus=True,
                    right_minus=True,
                ),
                content=None,
            ),
        ],
        closing_tag=None,
    )

    assert str(jinja.parse('{%- foo -%}')) == '{%- foo -%}'
    assert str(jinja.parse('{%- foo %}')) == '{%- foo %}'
    assert str(jinja.parse('{{- bar -}}')) == '{{- bar -}}'
    assert str(jinja.parse('{{ bar -}}')) == '{{ bar -}}'
    assert str(jinja.parse('{%+ foo %}')) == '{%+ foo %}'
    assert str(jinja.parse('{{+ bar }}')) == '{{+ bar }}'


def test_doctype():
    assert content.parse('<!DOCTYPE html>') == Interp('<!DOCTYPE html>')


def test_attrs():
    attrs = make_attributes_parser({}, jinja)
    parse = attrs.parse

    assert str(parse('{% if %}{% endif %}')) == '{% if %}{% endif %}'
    assert str(parse('{% if %}  {% endif %}')) == '{% if %}{% endif %}'
    assert str(parse('{% if %}a=b{% endif %}')) == '{% if %}a=b{% endif %}'
    assert str(parse('{% if %} a=b {% endif %}')) == '{% if %}a=b{% endif %}'


def test_optional_container():
    src = '{% if a %}<a href="b">{% endif %}c<b>d</b>{% if a %}</a>{% endif %}'
    assert src == str(content.parse(src))

    src = '''
    {% if a %} <a href="b"> {% endif %}
        c <b> d </b>
    {% if a %} </a> {% endif %}
    '''
    content.parse(src)


def test_whole_document():
    src = '<html lang="fr"><body>Hello<br></body></html>'
    assert src == str(element.parse(src))


def test():
    test_dummy_location()
    test_tag_name()
    test_attribute_value()
    test_attribute()
    test_comment()
    test_jinja_comment()
    test_opening_tag()
    test_closing_tag()
    test_raw_text_elements()
    test_element()
    test_self_closing_elements()
    test_jinja_blocks()
    test_jinja_whitespace_controls()
    test_doctype()
    test_attrs()
    test_optional_container()



import re

import parsy as P

from .ast import (
    Attribute, ClosingTag, Comment, Element, Integer, Interpolated,
    Jinja, JinjaComment, JinjaElement, JinjaElementPart, JinjaTag,
    JinjaVariable, JinjaOptionalContainer, Location, OpeningTag, String,
)
from .util import flatten


# TODO: Move it elsewhere
import sys
sys.setrecursionlimit(10000)


# Also called “void elements”
SELF_CLOSING_ELEMENTS = """
area base br col command embed hr img input keygen link meta param source
track wbr
""".split()

DEPRECATED_ELEMENTS = """
acronym applet basefont big center dir font frame frameset noframes isindex
strike tt
""".split()

DEFAULT_JINJA_STRUCTURED_ELEMENTS_NAMES = [
    ('autoescape', 'endautoescape'),
    ('block', 'endblock'),
    ('blocktrans', 'plural', 'endblocktrans'),
    ('comment', 'endcomment'),
    ('filter', 'endfilter'),
    ('for', 'else', 'empty', 'endfor'),
    ('if', 'elif', 'else', 'endif'),
    ('ifchanged', 'else', 'endifchanged'),
    ('ifequal', 'endifequal'),
    ('ifnotequal', 'endifnotequal'),
    ('spaceless', 'endspaceless'),
    ('verbatim', 'endverbatim'),
    ('with', 'endwith'),
]


# `\s` in an `str` pattern does match Unicode fancy spaces (especially the
# non-breaking ones). We’ll warn against some of these fancy spaces in the
# check phase.
# XXX: It could be better and simpler to only allow ASCII whitespaces here.
whitespace = P.regex(r'\s*')
mandatory_whitespace = P.regex(r'\s+')


def until(parser):
    return parser.should_fail(repr(parser)).then(P.any_char).many()


def combine_locations(begin_index, begin, result, end_index, end):
    begin_loc = Location(line=begin[0], column=begin[1], index=begin_index)
    end_loc = Location(line=end[0], column=end[1], index=end_index)
    locations = {
        'begin': begin_loc,
        'end': end_loc,
    }

    return (
        locations,
        result,
    )


def locate(parser):

    return (
        P.seq(
            P.index,
            P.line_info,
            parser,
            P.index,
            P.line_info,
        )
        .combine(combine_locations)
    )


def _combine_jinja_tag(locations, props):
    return JinjaTag(
        name=props['name'],
        content=props['extra_content'],
        left_plus=props['left_plus'],
        left_minus=props['left_minus'],
        right_minus=props['right_minus'],
        **locations,
    )


def _combine_jinja_variable(locations, props):
    return JinjaVariable(
        content=props['extra_content'],
        left_plus=props['left_plus'],
        left_minus=props['left_minus'],
        right_minus=props['right_minus'],
        **locations,
    )


def _combine_jinja_comment(locations, text):
    return JinjaComment(
        text=text,
        **locations,
    )


def _combine_jinja_tag_like(locations, props):
    return (
        locations,
        {
            'left_plus': props[0] is not None,
            'left_minus': props[1] is not None,
            'name': props[2],
            'extra_content': props[3],
            'right_minus': props[4] is not None,
        }
    )


def make_jinja_tag_like_parser(name, ml='{', mr='}'):
    """
    Create parsers for Jinja variables and regular Jinja tags.

    `name` should be a parser to parse the tag name.
    """
    end = whitespace.then(P.string('-').optional()).skip(P.string(mr + '}'))
    return locate(P.seq(
        P.string('{' + ml).then(P.string('+').optional()),
        P.string('-').optional().skip(whitespace),
        name.skip(whitespace),
        until(end).concat(),
        end
    )).combine(_combine_jinja_tag_like)


jinja_variable = make_jinja_tag_like_parser(
    P.success(None), '{', '}'
).combine(_combine_jinja_variable)


jinja_comment = (
    locate(
        P.string('{#')
        .skip(whitespace)
        .then(until(whitespace + P.string('#}')).concat())
        .skip(whitespace + P.string('#}'))
    )
    .combine(_combine_jinja_comment)
)


def make_jinja_tag_parser(name_parser):
    return (
        make_jinja_tag_like_parser(name_parser, '%', '%')
        .combine(_combine_jinja_tag)
    )


def _combine_jinja_element(locations, content):
    parts = list(flatten(content[0]))
    closing_tag = content[1] if len(content) == 2 else None

    e = JinjaElement(
        parts=parts,
        closing_tag=closing_tag,
        **locations,
    )
    return e


def _combine_jinja_element_part(locations, props):
    tag, content = props
    return JinjaElementPart(
        tag=tag,
        content=content,
        **locations,
    )


def make_jinja_element_part_parser(name_parser, content):
    return locate(P.seq(
        make_jinja_tag_parser(name_parser),
        content,
    )).combine(_combine_jinja_element_part)


def make_jinja_element_parser(name_parsers, content):
    """
    `name_parsers` must be a list of tag name parsers. For example,
    `name_parsers` can be defined as follow in order to parse `if` statements:

        name_parsers = [P.string(n) for n in ['if', 'elif', 'else', 'endif']]
    """

    if len(name_parsers) == 1:
        tag = make_jinja_tag_parser(name_parsers[0])
        part = locate(P.seq(
            tag, P.success(None),
        )).combine(_combine_jinja_element_part)
        parts = [part]
        end_tag_parser = None
    else:
        part_names = name_parsers[:-1]
        first_part = make_jinja_element_part_parser(
            part_names[0], content=content)
        next_parts = [
            make_jinja_element_part_parser(name, content=content).many()
            for name in part_names[1:]
        ]
        parts = [first_part] + next_parts
        end_tag_parser = make_jinja_tag_parser(name_parsers[-1])

    content = [P.seq(*parts)]
    if end_tag_parser:
        content.append(end_tag_parser)

    return (
        locate(P.seq(*content))
        .combine(_combine_jinja_element)
    )


jinja_name = P.letter + (
    P.letter | P.decimal_digit | P.string('_')
).many().concat()


def interpolated(parser):
    def combine_interpolated(locations, result):
        return Interpolated(
            nodes=result,
            **locations,
        )

    return (
        locate(parser)
        .combine(combine_interpolated)
    )


tag_name_start_char = P.regex(r'[:a-z]')
tag_name_char = tag_name_start_char | P.regex(r'[0-9-_.]')
tag_name = tag_name_start_char + tag_name_char.many().concat()

dtd = P.regex(r'<![^>]*>')

string_attribute_char = P.char_from('-_./+,?=:;#') | P.regex(r'[0-9a-zA-Z]')


def make_quoted_string_attribute_parser(quote, jinja):
    """
    quote: A single or a double quote
    """
    def combine(locations, value):
        return String(
            value=value,
            quote=quote,
            **locations,
        )

    value_char = P.regex(r'[^<]', flags=re.DOTALL)
    value = interpolated(
        P.string(quote).should_fail('no ' + quote)
        .then(jinja | value_char)
        .many()
    )

    return locate(
        P.string(quote)
        .then(value)
        .skip(P.string(quote))
    ).combine(combine)


def _combine_string_attribute_value(locations, value):
    return String(
        value=value,
        quote=None,
        **locations,
    )


def _combine_int_attribute_value(locations, value):
    return Integer(
        value=value,
        has_percent=False,
        **locations,
    )


int_attribute_value = locate(
    P.regex(r'[0-9]+')
    .map(int)
).combine(_combine_int_attribute_value)


def _combine_attribute(locations, props):
    name, equal_and_value = props
    value = None if equal_and_value is None else equal_and_value['value']
    return Attribute(
        name=name,
        value=value,
        **locations,
    )


def make_attribute_value_parser(jinja):
    string_attribute_value = locate(
        interpolated(
            (jinja | string_attribute_char)
            .at_least(1)
        )
    ).combine(_combine_string_attribute_value)

    return (
        make_quoted_string_attribute_parser('"', jinja) |
        make_quoted_string_attribute_parser("'", jinja) |
        int_attribute_value |
        string_attribute_value
    ).desc('attribute value')


def make_attribute_parser(jinja):
    attribute_value = make_attribute_value_parser(jinja)
    return (
        locate(
            P.seq(
                interpolated(tag_name),
                whitespace.then(
                    P.seq(
                        P.string('=').skip(whitespace).tag('equal'),
                        interpolated(attribute_value).tag('value'),
                    ).map(dict)
                ).optional(),
            )
        )
        .combine(_combine_attribute)
        .desc('attribute')
    )


def make_attributes_parser(config, jinja):
    attribute = make_attribute_parser(jinja)

    jinja_attr = make_jinja_parser(
        config,
        interpolated(
            whitespace
            .then(
                (attribute | jinja).sep_by(whitespace)
            )
            .skip(whitespace)
        )
    )

    attrs = interpolated(
        (
            whitespace.then(jinja_attr) |
            mandatory_whitespace.then(attribute)
        ).many()
    )

    return attrs


def _combine_comment(locations, text):
    return Comment(
        text=text,
        **locations,
    )


comment = locate(
    P.string('<!--')
    .then(P.regex(r'.*?(?=-->)', flags=re.DOTALL))
    .skip(P.string('-->'))
).combine(_combine_comment)


def _combine_opening_tag(locations, props):
    _lt, tag_name, attributes, slash, _gt = props
    return OpeningTag(
        name=tag_name,
        attributes=attributes,
        slash=slash,
        **locations,
    )


def _combine_closing_tag(locations, name):
    return ClosingTag(
        name=name,
        **locations
    )


def make_closing_tag_parser(tag_name_parser):
    return locate(
        P.string('</')
        .then(tag_name_parser)
        .skip(P.string('>'))
    ).combine(_combine_closing_tag)


def _combine_slash(locations, _):
    return locations['begin']


def make_opening_tag_parser(config,
                            jinja,
                            tag_name_parser=None,
                            allow_slash=False):
    attributes = make_attributes_parser(config, jinja)

    if not tag_name_parser:
        tag_name_parser = tag_name | jinja

    if allow_slash:
        slash = (
            locate(
                P.string('/')
                .skip(whitespace)
            )
            .combine(_combine_slash)
            .optional()
        )
    else:
        slash = P.success(None)

    return (
        locate(P.seq(
            P.string('<'),
            tag_name_parser,
            attributes.skip(whitespace),
            slash,
            P.string('>'),
        ))
        .combine(_combine_opening_tag)
    )


def _combine_element(locations, props):
    opening_tag, content, closing_tag = props
    return Element(
        opening_tag=opening_tag,
        closing_tag=closing_tag,
        content=content,
        **locations,
    )


def make_raw_text_element_parser(config, tag_name, jinja):
    """
    Used for <style> and <script>.
    """
    opening_tag = make_opening_tag_parser(
        config,
        tag_name_parser=P.string(tag_name),
        jinja=jinja,
    )

    body = P.regex(r'.*?(?=</' + tag_name + '>)', flags=re.DOTALL)

    closing_tag = make_closing_tag_parser(P.string(tag_name))

    return (
        locate(P.seq(
            opening_tag,
            body,
            closing_tag,
        ))
        .combine(_combine_element)
    )


slow_text_char = (
    (P.string('<') | P.string('{%') | P.string('{#') | P.string('{{'))
    .should_fail('text')
    .then(P.any_char)
)

# Not as precise as slow_text_char but must faster
quick_text = P.regex(r'[^{<]+', flags=re.DOTALL)


def _combine_optional_container(locations, nodes):
    return JinjaOptionalContainer(
        first_opening_if=nodes[0],
        opening_tag=nodes[1],
        first_closing_if=nodes[2],
        content=nodes[3],
        second_opening_if=nodes[4],
        closing_tag=nodes[5],
        second_closing_if=nodes[6],
        **locations,
    )


# Awkward hack to handle optional HTML containers, for example:
#
#   {% if a %}
#     <div>
#   {% endif %}
#   foo
#   {% if a %}
#     </div>
#   {% endif %}
#
# Currently, this only works with `if` statements and the two conditions
# must be exactly the same.
def make_jinja_optional_container_parser(config, content, jinja):
    jinja_if = make_jinja_tag_parser(P.string('if'))
    jinja_endif = make_jinja_tag_parser(P.string('endif'))
    opening_tag = make_opening_tag_parser(config, jinja, allow_slash=False)

    @P.generate
    def opt_container_impl():
        o_first_if_node = yield jinja_if.skip(whitespace)
        o_tag_node = yield opening_tag.skip(whitespace)
        c_first_if_node = yield jinja_endif

        content_nodes = yield content

        o_second_if_node = yield jinja_if.skip(whitespace)
        if o_second_if_node.content != o_first_if_node.content:
            yield P.fail('expected `{% if ' + content + ' %}`')
            return
        html_tag_name = o_tag_node.name
        if isinstance(html_tag_name, str):
            closing_tag = make_closing_tag_parser(P.string(html_tag_name))
        else:
            assert isinstance(html_tag_name, Jinja)
            closing_tag = make_closing_tag_parser(jinja)
        c_tag_node = yield closing_tag.skip(whitespace)
        c_second_if_node = yield jinja_endif

        return [
            o_first_if_node,
            o_tag_node,
            c_first_if_node,
            content_nodes,
            o_second_if_node,
            c_tag_node,
            c_second_if_node,
        ]

    return (
        locate(opt_container_impl)
        .combine(_combine_optional_container)
    )


def make_jinja_parser(config, content):
    # Allow to override elements with the configuration
    jinja_structured_elements_names = dict(
        (names[0], names)
        for names
        in (
            DEFAULT_JINJA_STRUCTURED_ELEMENTS_NAMES +
            config.get('jinja_custom_elements_names', [])
        )
    ).values()

    jinja_structured_element = P.alt(*[
        make_jinja_element_parser(
            [P.string(name) for name in names],
            content=content,
        )
        for names in jinja_structured_elements_names
    ])

    # These tag names can't begin a Jinja element
    jinja_intermediate_tag_names = set(
        n
        for _, *sublist in jinja_structured_elements_names
        for n in sublist
    )

    jinja_intermediate_tag_name = P.alt(*(
        P.string(n) for n in jinja_intermediate_tag_names
    ))

    jinja_element_single = make_jinja_element_parser(
        [
            P.alt(
                # HACK: If we allow `{% if %}`s without `{% endif %}`s here,
                # `make_jinja_optional_container_parser` doesn’t work. It
                # is probably better to reject any structured tag name here.
                P.string('if'),

                jinja_intermediate_tag_name
            )
            .should_fail('not an intermediate Jinja tag name')
            .then(jinja_name)
        ],
        content=content,
    )

    jinja_element = jinja_structured_element | jinja_element_single

    return jinja_variable | jinja_comment | jinja_element


def make_container_element_parser(config, content, jinja):
    opening_tag = make_opening_tag_parser(config, jinja=jinja)

    @P.generate
    def container_element_impl():
        o_tag_node = yield opening_tag
        content_nodes = yield content
        tag_name = o_tag_node.name
        if isinstance(tag_name, str):
            closing_tag = make_closing_tag_parser(P.string(tag_name))
        else:
            assert isinstance(tag_name, Jinja)
            closing_tag = make_closing_tag_parser(jinja)
        c_tag_node = yield closing_tag
        return [o_tag_node, content_nodes, c_tag_node]

    return (
        locate(container_element_impl)
        .combine(_combine_element)
    )


def make_element_parser(config, content, jinja):
    container_element = make_container_element_parser(
        config,
        content=content,
        jinja=jinja,
    )

    self_closing_element_opening_tag = make_opening_tag_parser(
        config,
        tag_name_parser=P.string_from(*SELF_CLOSING_ELEMENTS),
        allow_slash=True,
        jinja=jinja,
    )

    self_closing_element = (
        locate(P.seq(
            self_closing_element_opening_tag.skip(whitespace),
            P.success(None),  # No content
            P.success(None),  # No closing tag
        ))
        .combine(_combine_element)
    )

    style = make_raw_text_element_parser(config, 'style', jinja=jinja)
    script = make_raw_text_element_parser(config, 'script', jinja=jinja)

    return style | script | self_closing_element | container_element


def make_parser(config=None):
    if config is None:
        config = {}

    content_ = None

    @P.generate
    def content():
        c = yield content_
        return c

    jinja = make_jinja_parser(config, content)

    element = make_element_parser(config, content, jinja)

    opt_container = make_jinja_optional_container_parser(
        config, content, jinja)

    content_ = interpolated(
        (
            quick_text | comment | dtd | element | opt_container | jinja |
            slow_text_char
        ).many()
    )

    return {
        'content': content_,
        'jinja': jinja,
        'element': element,
    }



import tempfile

from .config import parse_config


def test():
    with tempfile.NamedTemporaryFile() as f:
        f.write(b'hello = "worl" + "d"\n')
        f.seek(0)
        config = parse_config(f.name)
        assert config == {'hello': 'world'}



from .cli import main

__name__ = "curlylint"
__version__ = "0.5.0"
__description__ = "A linter for Jinja-like templates. Forked from jinjalint"
__author__ = "Thibaud Colas"
__author_email__ = "thibaudcolas@gmail.com"
__url__ = "https://github.com/thibaudcolas/curlylint"
__license__ = "MIT"
__copyright__ = "Copyright 2020-present Thibaud Colas"

if __name__ == '__main__':
    main()



from pathlib import Path


def parse_config(path):
    path = Path(path)
    with path.open('r') as f:
        source = f.read()
    config = {}
    exec(source, config)
    config = dict(config)
    del config['__builtins__']
    return config



import attr

from . import ast
from .file import File


@attr.s(frozen=True)
class IssueLocation:
    file_path = attr.ib()  # Path
    column = attr.ib()
    line = attr.ib()

    def __str__(self):
        return '{}:{}:{}'.format(
            self.file_path,
            self.line + 1,
            self.column,
        )

    @staticmethod
    def from_ast(file_path, ast_location):
        if isinstance(file_path, File):
            file_path = file_path.path

        return IssueLocation(
            file_path=file_path,
            column=ast_location.column,
            line=ast_location.line,
        )


@attr.s(frozen=True)
class Issue:
    location = attr.ib()
    message = attr.ib()

    def __str__(self):
        return '{}: {}'.format(self.location, self.message)

    def __attrs_post_init__(self):
        assert isinstance(self.location, IssueLocation)

    @staticmethod
    def from_ast(file_path, ast_location, message):
        return Issue(
            IssueLocation.from_ast(file_path, ast_location),
            message,
        )



from .util import flatten


def test():
    assert list(flatten(())) == []
    assert list(flatten([])) == []
    assert list(flatten((1,))) == [1]
    assert list(flatten([2, [], (), [3, [(4, 5), (6,)]]])) == [2, 3, 4, 5, 6]



import collections


def flatten(l):
    """
    Deeply flattens an iterable.
    """
    for el in l:
        if (isinstance(el, collections.Iterable) and
                not isinstance(el, (str, bytes))):
            yield from flatten(el)
        else:
            yield el



"""curlylint

Usage:
  curlylint [-v | --verbose] [--config CONFIG] [--parse-only]
            [--extension EXT | -e EXT]... [INPUT ...]
  curlylint (-h | --help)
  curlylint --version

Options:
  -h --help             Show this help message and exit.
  --version             Show version information and exit.
  -v --verbose          Verbose mode.
  -c --config CONFIG    Specify the configuration file.
  --parse-only          Don’t lint, check for syntax errors and exit.
  -e --extension EXT    Extension of the files to analyze (used if INPUT
                        contains directories to crawl).
                        [default: html jinja twig]

The configuration file must be a valid Python file.
"""
from docopt import docopt

from .lint import lint, resolve_file_paths
from .config import parse_config


def print_issues(issues, config):
    sorted_issues = sorted(
        issues,
        key=lambda i: (
            i.location.file_path,
            i.location.line,
            i.location.column
        ),
    )

    for issue in sorted_issues:
        print(str(issue))


def main():
    arguments = docopt(__doc__)

    input_names = arguments['INPUT'] or ['.']
    extensions = ['.' + e for e in arguments['--extension']]
    verbose = arguments['--verbose']

    if arguments['--version']:
        print('0.5.0')
        return

    if arguments['--config']:
        if verbose:
            print('Using configuration file {}'.format(arguments['--config']))
        config = parse_config(arguments['--config'])
    else:
        config = {}

    config['verbose'] = verbose
    config['parse_only'] = arguments['--parse-only']

    paths = list(resolve_file_paths(input_names, extensions=extensions))

    if verbose:
        print('Files being analyzed:')
        print('\n'.join(str(p) for p in paths))
        print()

    issues = lint(paths, config)
    print_issues(issues, config)

    if any(issues):
        exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import io

from setuptools import find_packages, setup

from curlylint import (
    __author__,
    __author_email__,
    __copyright__,
    __description__,
    __license__,
    __name__,
    __url__,
    __version__,
)

with io.open('README.md', encoding='utf-8') as readme_file:
    long_description = readme_file.read()

setup(
    name=__name__,
    version=__version__,
    description=__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=__url__,
    author=__author__,
    author_email=__author_email__,
    license=__license__,
    copyright=__copyright__,
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    keywords=[
        "curlylint",
        "jinjalint",
        "linter",
        "jinja",
        "jinja2",
        "django templates",
        "templates",
        "html",
        "twig",
    ],
    install_requires=[
        'parsy==1.1.0',
        'attrs==17.2.0',
        'docopt==0.6.2',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Quality Assurance',
        'Environment :: Console',
        'Framework :: Django',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    entry_points={
        'console_scripts': ['curlylint=curlylint.cli:main'],
    },
)
